@@ -41,6 +41,6 @@ No OS subroutines are required.
 int
 _DEFUN(isalnum,(c),int c)
 {
-	return((_ctype_ + 1)[c] & (_U|_L|_N));
+	return(__ctype_ptr__[c+1] & (_U|_L|_N));
 }
 
