@@ -37,10 +37,14 @@
 #include "lmonetary.h"
 #include "lmessages.h"
 
+#ifndef __CYGWIN__
 #define TRANSITION_PERIOD_HACK
+#endif
 
 #define _REL(BASE) ((int)item-BASE)
 
+extern char *__locale_charset ();
+
 char *
 _DEFUN(nl_langinfo, (item), 
        nl_item item) {
@@ -54,6 +58,15 @@ _DEFUN(nl_langinfo, (item),
 
    switch (item) {
 	case CODESET:
+#ifdef __CYGWIN__
+		ret = __locale_charset ();
+		/* Temporary exception for KOI8 charsets which are
+		   incorrectly treated by calling applications otherwise. */
+		if (strcmp (ret, "CP20866") == 0)
+		  ret = "KOI8-R";
+		else if (strcmp (ret, "CP21866") == 0)
+		  ret = "KOI8-U";
+#else
 		ret = "";
 		if ((s = setlocale(LC_CTYPE, NULL)) != NULL) {
 			if ((cs = strchr(s, '.')) != NULL) {
@@ -92,6 +105,7 @@ _DEFUN(nl_langinfo, (item),
 				  )
 				ret = "US-ASCII";
 		}
+#endif /* __CYGWIN__ */
 		break;
 	case D_T_FMT:
 		ret = (char *) __get_current_time_locale()->c_fmt;
