@@ -21,11 +21,16 @@ FUNCTION
 
 INDEX
 	fgets
+INDEX
+	_fgets_r
 
 ANSI_SYNOPSIS
         #include <stdio.h>
 	char *fgets(char *<[buf]>, int <[n]>, FILE *<[fp]>);
 
+        #include <stdio.h>
+	char *_fgets_r(struct _reent *<[ptr]>, char *<[buf]>, int <[n]>, FILE *<[fp]>);
+
 TRAD_SYNOPSIS
 	#include <stdio.h>
 	char *fgets(<[buf]>,<[n]>,<[fp]>)
@@ -33,11 +38,21 @@ TRAD_SYNOPSIS
 	int <[n]>;
 	FILE *<[fp]>;
 
+	#include <stdio.h>
+	char *_fgets_r(<[ptr]>, <[buf]>,<[n]>,<[fp]>)
+	struct _reent *<[ptr]>;
+        char *<[buf]>;
+	int <[n]>;
+	FILE *<[fp]>;
+
 DESCRIPTION
 	Reads at most <[n-1]> characters from <[fp]> until a newline
 	is found. The characters including to the newline are stored
 	in <[buf]>. The buffer is terminated with a 0.
 
+	The <<_fgets_r>> function is simply the reentrant version of
+	<<fgets>> and is passed an additional reentrancy structure
+	pointer: <[ptr]>.
 
 RETURNS
 	<<fgets>> returns the buffer passed to it, with the data
@@ -66,7 +81,8 @@ Supporting OS subroutines required: <<close>>, <<fstat>>, <<isatty>>,
  */
 
 char *
-_DEFUN(fgets, (buf, n, fp),
+_DEFUN(_fgets_r, (ptr, buf, n, fp),
+       struct _reent * ptr _AND
        char *buf _AND
        int n     _AND
        FILE * fp)
@@ -80,15 +96,16 @@ _DEFUN(fgets, (buf, n, fp),
 
   s = buf;
 
-  CHECK_INIT(_REENT);
+  CHECK_INIT(ptr, fp);
 
+  __sfp_lock_acquire ();
   _flockfile (fp);
 #ifdef __SCLE
   if (fp->_flags & __SCLE)
     {
       int c;
       /* Sorry, have to do it the slow way */
-      while (--n > 0 && (c = __sgetc (fp)) != EOF)
+      while (--n > 0 && (c = __sgetc_r (ptr, fp)) != EOF)
 	{
 	  *s++ = c;
 	  if (c == '\n')
@@ -97,10 +114,12 @@ _DEFUN(fgets, (buf, n, fp),
       if (c == EOF && s == buf)
         {
           _funlockfile (fp);
+	  __sfp_lock_release ();
           return NULL;
         }
       *s = 0;
       _funlockfile (fp);
+      __sfp_lock_release ();
       return buf;
     }
 #endif
@@ -113,12 +132,13 @@ _DEFUN(fgets, (buf, n, fp),
        */
       if ((len = fp->_r) <= 0)
 	{
-	  if (__srefill (fp))
+	  if (__srefill_r (ptr, fp))
 	    {
 	      /* EOF: stop with partial or no line */
 	      if (s == buf)
                 {
                   _funlockfile (fp);
+		  __sfp_lock_release ();
                   return 0;
                 }
 	      break;
@@ -144,6 +164,7 @@ _DEFUN(fgets, (buf, n, fp),
 	  _CAST_VOID memcpy ((_PTR) s, (_PTR) p, len);
 	  s[len] = 0;
           _funlockfile (fp);
+	  __sfp_lock_release ();
 	  return (buf);
 	}
       fp->_r -= len;
@@ -154,5 +175,19 @@ _DEFUN(fgets, (buf, n, fp),
   while ((n -= len) != 0);
   *s = 0;
   _funlockfile (fp);
+  __sfp_lock_release ();
   return buf;
 }
+
+#ifndef _REENT_ONLY
+
+char *
+_DEFUN(fgets, (buf, n, fp),
+       char *buf _AND
+       int n     _AND
+       FILE * fp)
+{
+  return _fgets_r (_REENT, buf, n, fp);
+}
+
+#endif /* !_REENT_ONLY */
