@@ -23,56 +23,57 @@
 
 /*
 FUNCTION
-<<mktemp>>, <<mkstemp>>---generate unused file name
+<<mktemp>>, <<mkstemp>>, <<mkstemps>>---generate unused file name
+<<mkdtemp>>---generate unused directory
 
 INDEX
 	mktemp
+INDEX
+	mkdtemp
 INDEX
 	mkstemp
+INDEX
+	mkstemps
 INDEX
 	_mktemp_r
+INDEX
+	_mkdtemp_r
 INDEX
 	_mkstemp_r
+INDEX
+	_mkstemps_r
 
 ANSI_SYNOPSIS
-	#include <stdio.h>
+	#include <stdlib.h>
 	char *mktemp(char *<[path]>);
+	char *mkdtemp(char *<[path]>);
 	int mkstemp(char *<[path]>);
+	int mkstemps(char *<[path]>, int <[suffixlen]>);
 
 	char *_mktemp_r(struct _reent *<[reent]>, char *<[path]>);
+	char *_mkdtemp_r(struct _reent *<[reent]>, char *<[path]>);
 	int *_mkstemp_r(struct _reent *<[reent]>, char *<[path]>);
-
-TRAD_SYNOPSIS
-	#include <stdio.h>
-	char *mktemp(<[path]>)
-	char *<[path]>;
-
-	int mkstemp(<[path]>)
-	char *<[path]>;
-
-	char *_mktemp_r(<[reent]>, <[path]>)
-	struct _reent *<[reent]>;
-	char *<[path]>;
-
-	int _mkstemp_r(<[reent]>, <[path]>)
-	struct _reent *<[reent]>;
-	char *<[path]>;
+	int *_mkstemps_r(struct _reent *<[reent]>, char *<[path]>, int <[len]>);
 
 DESCRIPTION
-<<mktemp>> and <<mkstemp>> attempt to generate a file name that is not
-yet in use for any existing file.  <<mkstemp>> creates the file and 
-opens it for reading and writing; <<mktemp>> simply generates the file name.
+<<mktemp>>, <<mkstemp>>, and <<mkstemps>> attempt to generate a file name
+that is not yet in use for any existing file.  <<mkstemp>> and <<mkstemps>>
+create the file and open it for reading and writing; <<mktemp>> simply
+generates the file name (making <<mktemp>> a security risk).  <<mkdtemp>>
+attempts to create a directory instead of a file, with a permissions
+mask of 0700.
 
 You supply a simple pattern for the generated file name, as the string
 at <[path]>.  The pattern should be a valid filename (including path
-information if you wish) ending with some number of `<<X>>'
+information if you wish) ending with at least six `<<X>>'
 characters.  The generated filename will match the leading part of the
 name you supply, with the trailing `<<X>>' characters replaced by some
-combination of digits and letters.
+combination of digits and letters.  With <<mkstemps>>, the `<<X>>'
+characters end <[suffixlen]> bytes before the end of the string.
 
-The alternate functions <<_mktemp_r>> and <<_mkstemp_r>> are reentrant
-versions.  The extra argument <[reent]> is a pointer to a reentrancy
-structure.
+The alternate functions <<_mktemp_r>>, <<_mkdtemp_r>>, <<_mkstemp_r>>,
+and <<_mkstemps_r>> are reentrant versions.  The extra argument <[reent]>
+is a pointer to a reentrancy structure.
 
 RETURNS
 <<mktemp>> returns the pointer <[path]> to the modified string
@@ -80,19 +81,32 @@ representing an unused filename, unless it could not generate one, or
 the pattern you provided is not suitable for a filename; in that case,
 it returns <<NULL>>.
 
-<<mkstemp>> returns a file descriptor to the newly created file,
-unless it could not generate an unused filename, or the pattern you
+<<mkdtemp>> returns the pointer <[path]> to the modified string if the
+directory was created, otherwise it returns <<NULL>>.
+
+<<mkstemp>> and <<mkstemps>> return a file descriptor to the newly created
+file, unless it could not generate an unused filename, or the pattern you
 provided is not suitable for a filename; in that case, it returns
 <<-1>>.
 
+NOTES
+Never use <<mktemp>>.  The generated filenames are easy to guess and
+there's a race between the test if the file exists and the creation
+of the file.  In combination this makes <<mktemp>> prone to attacks
+and using it is a security risk.  Whenever possible use <<mkstemp>>
+instead.  It doesn't suffer the race condition.
+
 PORTABILITY
 ANSI C does not require either <<mktemp>> or <<mkstemp>>; the System
-V Interface Definition requires <<mktemp>> as of Issue 2.
+V Interface Definition requires <<mktemp>> as of Issue 2.  POSIX 2001
+requires <<mkstemp>>, and POSIX 2008 requires <<mkdtemp>>, but
+<<mkstemps>> is not standardized.
 
-Supporting OS subroutines required: <<getpid>>, <<open>>, <<stat>>.
+Supporting OS subroutines required: <<getpid>>, <<mkdir>>, <<open>>, <<stat>>.
 */
 
 #include <_ansi.h>
+#include <stdlib.h>
 #include <reent.h>
 #include <sys/types.h>
 #include <fcntl.h>
@@ -102,23 +116,42 @@ Supporting OS subroutines required: <<getpid>>, <<open>>, <<stat>>.
 #include <ctype.h>
 
 static int
-_DEFUN(_gettemp, (ptr, path, doopen),
+_DEFUN(_gettemp, (ptr, path, doopen, domkdir, suffixlen),
        struct _reent *ptr _AND
        char *path         _AND
-       register int *doopen)
+       register int *doopen _AND
+       int domkdir        _AND
+       size_t suffixlen)
 {
   register char *start, *trv;
+  char *end;
+#ifdef __USE_INTERNAL_STAT64
+  struct stat64 sbuf;
+#else
   struct stat sbuf;
+#endif
   unsigned int pid;
 
   pid = _getpid_r (ptr);
   for (trv = path; *trv; ++trv)		/* extra X's get set to 0's */
     continue;
-  while (*--trv == 'X')
+  if (trv - path < suffixlen)
+    {
+      ptr->_errno = EINVAL;
+      return 0;
+    }
+  trv -= suffixlen;
+  end = trv;
+  while (path < trv && *--trv == 'X')
     {
       *trv = (pid % 10) + '0';
       pid /= 10;
     }
+  if (end - trv < 6)
+    {
+      ptr->_errno = EINVAL;
+      return 0;
+    }
 
   /*
    * Check the target directory; if you have six X's and it
@@ -150,16 +183,27 @@ _DEFUN(_gettemp, (ptr, path, doopen),
 
   for (;;)
     {
+#if !defined _ELIX_LEVEL || _ELIX_LEVEL >= 4
+      if (domkdir)
+	{
+#ifdef HAVE_MKDIR
+	  if (_mkdir_r (ptr, path, 0700) == 0)
+	    return 1;
+	  if (ptr->_errno != EEXIST)
+	    return 0;
+#else /* !HAVE_MKDIR */
+	  ptr->_errno = ENOSYS;
+	  return 0;
+#endif /* !HAVE_MKDIR */
+	}
+      else
+#endif /* _ELIX_LEVEL */
       if (doopen)
 	{
 	  if ((*doopen = _open_r (ptr, path, O_CREAT | O_EXCL | O_RDWR, 0600))
 	      >= 0)
 	    return 1;
-#if defined(__CYGWIN__)
-	  if (ptr->_errno != EEXIST && ptr->_errno != EACCES)
-#else
 	  if (ptr->_errno != EEXIST)
-#endif
 	    return 0;
 	}
 #ifdef __USE_INTERNAL_STAT64
@@ -172,12 +216,13 @@ _DEFUN(_gettemp, (ptr, path, doopen),
       /* tricky little algorithm for backward compatibility */
       for (trv = start;;)
 	{
-	  if (!*trv)
+	  if (trv == end)
 	    return 0;
 	  if (*trv == 'z')
 	    *trv++ = 'a';
 	  else
 	    {
+	      /* Safe, since it only encounters 7-bit characters.  */
 	      if (isdigit (*trv))
 		*trv = 'a';
 	      else
@@ -196,15 +241,36 @@ _DEFUN(_mkstemp_r, (ptr, path),
 {
   int fd;
 
-  return (_gettemp (ptr, path, &fd) ? fd : -1);
+  return (_gettemp (ptr, path, &fd, 0, 0) ? fd : -1);
 }
 
+#if !defined _ELIX_LEVEL || _ELIX_LEVEL >= 4
+char *
+_DEFUN(_mkdtemp_r, (ptr, path),
+       struct _reent *ptr _AND
+       char *path)
+{
+  return (_gettemp (ptr, path, (int *) NULL, 1, 0) ? path : NULL);
+}
+
+int
+_DEFUN(_mkstemps_r, (ptr, path, len),
+       struct _reent *ptr _AND
+       char *path _AND
+       int len)
+{
+  int fd;
+
+  return (_gettemp (ptr, path, &fd, 0, len) ? fd : -1);
+}
+#endif /* _ELIX_LEVEL */
+
 char *
 _DEFUN(_mktemp_r, (ptr, path),
        struct _reent *ptr _AND
        char *path)
 {
-  return (_gettemp (ptr, path, (int *) NULL) ? path : (char *) NULL);
+  return (_gettemp (ptr, path, (int *) NULL, 0, 0) ? path : (char *) NULL);
 }
 
 #ifndef _REENT_ONLY
@@ -215,14 +281,33 @@ _DEFUN(mkstemp, (path),
 {
   int fd;
 
-  return (_gettemp (_REENT, path, &fd) ? fd : -1);
+  return (_gettemp (_REENT, path, &fd, 0, 0) ? fd : -1);
 }
 
+# if !defined _ELIX_LEVEL || _ELIX_LEVEL >= 4
+char *
+_DEFUN(mkdtemp, (path),
+       char *path)
+{
+  return (_gettemp (_REENT, path, (int *) NULL, 1, 0) ? path : NULL);
+}
+
+int
+_DEFUN(mkstemps, (path, len),
+       char *path _AND
+       int len)
+{
+  int fd;
+
+  return (_gettemp (_REENT, path, &fd, 0, len) ? fd : -1);
+}
+# endif /* _ELIX_LEVEL */
+
 char *
 _DEFUN(mktemp, (path),
        char *path)
 {
-  return (_gettemp (_REENT, path, (int *) NULL) ? path : (char *) NULL);
+  return (_gettemp (_REENT, path, (int *) NULL, 0, 0) ? path : (char *) NULL);
 }
 
 #endif /* ! defined (_REENT_ONLY) */
