@@ -19,6 +19,7 @@
 #include <_ansi.h>
 #include <stdio.h>
 #include <stdlib.h>
+#include <errno.h>
 #include "local.h"
 
 static int
@@ -36,12 +37,15 @@ _DEFUN(lflush, (fp),
  */
 
 int
-_DEFUN(__srefill, (fp),
+_DEFUN(__srefill_r, (ptr, fp),
+       struct _reent * ptr _AND
        register FILE * fp)
 {
   /* make sure stdio is set up */
 
-  CHECK_INIT (_REENT);
+  CHECK_INIT (ptr, fp);
+
+  ORIENT (fp, -1);
 
   fp->_r = 0;			/* largely a convenience for callers */
 
@@ -55,11 +59,15 @@ _DEFUN(__srefill, (fp),
   if ((fp->_flags & __SRD) == 0)
     {
       if ((fp->_flags & __SRW) == 0)
-	return EOF;
+	{
+	  ptr->_errno = EBADF;
+	  fp->_flags |= __SERR;
+	  return EOF;
+	}
       /* switch to reading */
       if (fp->_flags & __SWR)
 	{
-	  if (fflush (fp))
+	  if (_fflush_r (ptr, fp))
 	    return EOF;
 	  fp->_flags &= ~__SWR;
 	  fp->_w = 0;
@@ -77,7 +85,7 @@ _DEFUN(__srefill, (fp),
        */
       if (HASUB (fp))
 	{
-	  FREEUB (fp);
+	  FREEUB (ptr, fp);
 	  if ((fp->_r = fp->_ur) != 0)
 	    {
 	      fp->_p = fp->_up;
@@ -87,7 +95,7 @@ _DEFUN(__srefill, (fp),
     }
 
   if (fp->_bf._base == NULL)
-    __smakebuf (fp);
+    __smakebuf_r (ptr, fp);
 
   /*
    * Before reading from a line buffered or unbuffered file,
@@ -98,8 +106,7 @@ _DEFUN(__srefill, (fp),
   if (fp->_flags & (__SLBF | __SNBF))
     _CAST_VOID _fwalk (_GLOBAL_REENT, lflush);
   fp->_p = fp->_bf._base;
-  fp->_r = (*fp->_read) (fp->_cookie, (char *) fp->_p, fp->_bf._size);
-  fp->_flags &= ~__SMOD;	/* buffer contents are again pristine */
+  fp->_r = fp->_read (ptr, fp->_cookie, (char *) fp->_p, fp->_bf._size);
 #ifndef __CYGWIN__
   if (fp->_r <= 0)
 #else
