@@ -17,14 +17,20 @@
 
 /*
 FUNCTION
-<<vscanf>>, <<vfscanf>>, <<vsscanf>>---format argument list
+<<vfscanf>>, <<vscanf>>, <<vsscanf>>---format argument list
 
-INDEX
-	vscanf
 INDEX
 	vfscanf
+INDEX
+	_vfscanf_r
+INDEX
+	vscanf
+INDEX
+	_vscanf_r
 INDEX
 	vsscanf
+INDEX
+	_vsscanf_r
 
 ANSI_SYNOPSIS
 	#include <stdio.h>
@@ -33,9 +39,9 @@ ANSI_SYNOPSIS
 	int vfscanf(FILE *<[fp]>, const char *<[fmt]>, va_list <[list]>);
 	int vsscanf(const char *<[str]>, const char *<[fmt]>, va_list <[list]>);
 
-	int _vscanf_r(struct _reent *<[reent]>, const char *<[fmt]>, 
+	int _vscanf_r(struct _reent *<[reent]>, const char *<[fmt]>,
                        va_list <[list]>);
-	int _vfscanf_r(struct _reent *<[reent]>, FILE *<[fp]>, const char *<[fmt]>, 
+	int _vfscanf_r(struct _reent *<[reent]>, FILE *<[fp]>, const char *<[fmt]>,
                        va_list <[list]>);
 	int _vsscanf_r(struct _reent *<[reent]>, const char *<[str]>,
                        const char *<[fmt]>, va_list <[list]>);
@@ -51,7 +57,7 @@ TRAD_SYNOPSIS
 	FILE *<[fp]>;
 	char *<[fmt]>;
 	va_list <[list]>;
-	
+
 	int vsscanf( <[str]>, <[fmt]>, <[list]>)
 	char *<[str]>;
 	char *<[fmt]>;
@@ -67,7 +73,7 @@ TRAD_SYNOPSIS
 	FILE *<[fp]>;
 	char *<[fmt]>;
 	va_list <[list]>;
-	
+
 	int _vsscanf_r( <[reent]>, <[str]>, <[fmt]>, <[list]>)
 	struct _reent *<[reent]>;
 	char *<[str]>;
@@ -76,18 +82,18 @@ TRAD_SYNOPSIS
 
 DESCRIPTION
 <<vscanf>>, <<vfscanf>>, and <<vsscanf>> are (respectively) variants
-of <<scanf>>, <<fscanf>>, and <<sscanf>>.  They differ only in 
-allowing their caller to pass the variable argument list as a 
-<<va_list>> object (initialized by <<va_start>>) rather than 
+of <<scanf>>, <<fscanf>>, and <<sscanf>>.  They differ only in
+allowing their caller to pass the variable argument list as a
+<<va_list>> object (initialized by <<va_start>>) rather than
 directly accepting a variable number of arguments.
 
 RETURNS
 The return values are consistent with the corresponding functions:
 <<vscanf>> returns the number of input fields successfully scanned,
 converted, and stored; the return value does not include scanned
-fields which were not stored.  
+fields which were not stored.
 
-If <<vscanf>> attempts to read at end-of-file, the return value 
+If <<vscanf>> attempts to read at end-of-file, the return value
 is <<EOF>>.
 
 If no fields were stored, the return value is <<0>>.
@@ -109,31 +115,48 @@ Supporting OS subroutines required:
 #include <wctype.h>
 #include <stdio.h>
 #include <stdlib.h>
+#include <stdint.h>
 #include <limits.h>
 #include <wchar.h>
 #include <string.h>
-#ifdef _HAVE_STDC
 #include <stdarg.h>
-#else
-#include <varargs.h>
-#endif
+#include <errno.h>
 #include "local.h"
+#include "../stdlib/local.h"
 
 #ifdef INTEGER_ONLY
 #define VFSCANF vfiscanf
 #define _VFSCANF_R _vfiscanf_r
 #define __SVFSCANF __svfiscanf
-#define __SVFSCANF_R __svfiscanf_r
+#ifdef STRING_ONLY
+#  define __SVFSCANF_R __ssvfiscanf_r
+#else
+#  define __SVFSCANF_R __svfiscanf_r
+#endif
 #else
 #define VFSCANF vfscanf
 #define _VFSCANF_R _vfscanf_r
 #define __SVFSCANF __svfscanf
-#define __SVFSCANF_R __svfscanf_r
+#ifdef STRING_ONLY
+#  define __SVFSCANF_R __ssvfscanf_r
+#else
+#  define __SVFSCANF_R __svfscanf_r
+#endif
 #ifndef NO_FLOATING_POINT
 #define FLOATING_POINT
 #endif
 #endif
 
+#ifdef STRING_ONLY
+#undef _flockfile
+#undef _funlockfile
+#define _flockfile(x) {}
+#define _funlockfile(x) {}
+#define _ungetc_r _sungetc_r
+#define __srefill_r __ssrefill_r
+#define _fread_r _sfread_r
+#endif
+
 #ifdef FLOATING_POINT
 #include <math.h>
 #include <float.h>
@@ -163,10 +186,23 @@ extern _LONG_DOUBLE _strtold _PARAMS((char *s, char **sptr));
 #endif
 
 #define _NO_LONGLONG
-#if defined _WANT_IO_LONG_LONG && defined __GNUC__
+#if defined _WANT_IO_LONG_LONG \
+	&& (defined __GNUC__ || __STDC_VERSION__ >= 199901L)
 # undef _NO_LONGLONG
 #endif
 
+#define _NO_POS_ARGS
+#ifdef _WANT_IO_POS_ARGS
+# undef _NO_POS_ARGS
+# ifdef NL_ARGMAX
+#  define MAX_POS_ARGS NL_ARGMAX
+# else
+#  define MAX_POS_ARGS 32
+# endif
+
+static void * get_arg (int, va_list *, int *, void **);
+#endif /* _WANT_IO_POS_ARGS */
+
 /*
  * Flags used during conversion.
  */
@@ -215,20 +251,20 @@ extern _LONG_DOUBLE _strtold _PARAMS((char *s, char **sptr));
 typedef unsigned long long u_long_long;
 #endif
 
-/*static*/ u_char *__sccl ();
-
 /*
  * vfscanf
  */
 
 #define BufferEmpty (fp->_r <= 0 && __srefill_r(rptr, fp))
 
+#ifndef STRING_ONLY
+
 #ifndef _REENT_ONLY
 
 int
-_DEFUN(VFSCANF, (fp, fmt, ap), 
-       register FILE *fp _AND 
-       _CONST char *fmt _AND 
+_DEFUN(VFSCANF, (fp, fmt, ap),
+       register FILE *fp _AND
+       _CONST char *fmt _AND
        va_list ap)
 {
   CHECK_INIT(_REENT, fp);
@@ -248,15 +284,146 @@ _DEFUN(__SVFSCANF, (fp, fmt0, ap),
 
 int
 _DEFUN(_VFSCANF_R, (data, fp, fmt, ap),
-       struct _reent *data _AND 
-       register FILE *fp   _AND 
-       _CONST char *fmt    _AND 
+       struct _reent *data _AND
+       register FILE *fp   _AND
+       _CONST char *fmt    _AND
        va_list ap)
 {
   CHECK_INIT(data, fp);
   return __SVFSCANF_R (data, fp, fmt, ap);
 }
+#endif /* !STRING_ONLY */
 
+#if defined (STRING_ONLY) && defined (INTEGER_ONLY)
+/* When dealing with the sscanf family, we don't want to use the
+ * regular ungetc which will drag in file I/O items we don't need.
+ * So, we create our own trimmed-down version.  */
+int
+_DEFUN(_sungetc_r, (data, fp, ch),
+	struct _reent *data _AND
+	int c               _AND
+	register FILE *fp)
+{
+  if (c == EOF)
+    return (EOF);
+
+  /* After ungetc, we won't be at eof anymore */
+  fp->_flags &= ~__SEOF;
+  c = (unsigned char) c;
+
+  /*
+   * If we are in the middle of ungetc'ing, just continue.
+   * This may require expanding the current ungetc buffer.
+   */
+
+  if (HASUB (fp))
+    {
+      if (fp->_r >= fp->_ub._size && __submore (data, fp))
+        {
+          return EOF;
+        }
+      *--fp->_p = c;
+      fp->_r++;
+      return c;
+    }
+
+  /*
+   * If we can handle this by simply backing up, do so,
+   * but never replace the original character.
+   * (This makes sscanf() work when scanning `const' data.)
+   */
+
+  if (fp->_bf._base != NULL && fp->_p > fp->_bf._base && fp->_p[-1] == c)
+    {
+      fp->_p--;
+      fp->_r++;
+      return c;
+    }
+
+  /*
+   * Create an ungetc buffer.
+   * Initially, we will use the `reserve' buffer.
+   */
+
+  fp->_ur = fp->_r;
+  fp->_up = fp->_p;
+  fp->_ub._base = fp->_ubuf;
+  fp->_ub._size = sizeof (fp->_ubuf);
+  fp->_ubuf[sizeof (fp->_ubuf) - 1] = c;
+  fp->_p = &fp->_ubuf[sizeof (fp->_ubuf) - 1];
+  fp->_r = 1;
+  return c;
+}
+
+/* String only version of __srefill_r for sscanf family.  */
+int
+_DEFUN(__ssrefill_r, (ptr, fp),
+       struct _reent * ptr _AND
+       register FILE * fp)
+{
+  /*
+   * Our only hope of further input is the ungetc buffer.
+   * If there is anything in that buffer to read, return.
+   */
+  if (HASUB (fp))
+    {
+      FREEUB (ptr, fp);
+      if ((fp->_r = fp->_ur) != 0)
+        {
+          fp->_p = fp->_up;
+	  return 0;
+        }
+    }
+
+  /* Otherwise we are out of character input.  */
+  fp->_p = fp->_bf._base;
+  fp->_r = 0;
+  fp->_flags |= __SEOF;
+  return EOF;
+}
+
+size_t
+_DEFUN(_sfread_r, (ptr, buf, size, count, fp),
+       struct _reent * ptr _AND
+       _PTR buf _AND
+       size_t size _AND
+       size_t count _AND
+       FILE * fp)
+{
+  register size_t resid;
+  register char *p;
+  register int r;
+  size_t total;
+
+  if ((resid = count * size) == 0)
+    return 0;
+
+  total = resid;
+  p = buf;
+
+  while (resid > (r = fp->_r))
+    {
+      _CAST_VOID memcpy ((_PTR) p, (_PTR) fp->_p, (size_t) r);
+      fp->_p += r;
+      fp->_r = 0;
+      p += r;
+      resid -= r;
+      if (__ssrefill_r (ptr, fp))
+        {
+          /* no more input: return partial result */
+          return (total - resid) / size;
+        }
+    }
+  _CAST_VOID memcpy ((_PTR) p, (_PTR) fp->_p, resid);
+  fp->_r -= resid;
+  fp->_p += resid;
+  return count;
+}
+#else /* !STRING_ONLY || !INTEGER_ONLY */
+int _EXFUN (_sungetc_r, (struct _reent *, int, register FILE *));
+int _EXFUN (__ssrefill_r, (struct _reent *, register FILE *));
+size_t _EXFUN (_sfread_r, (struct _reent *, _PTR buf, size_t, size_t, FILE *));
+#endif /* !STRING_ONLY || !INTEGER_ONLY */
 
 int
 _DEFUN(__SVFSCANF_R, (rptr, fp, fmt0, ap),
@@ -274,6 +441,13 @@ _DEFUN(__SVFSCANF_R, (rptr, fp, fmt0, ap),
   register char *p0;		/* saves original value of p when necessary */
   int nassigned;		/* number of fields assigned */
   int nread;			/* number of characters consumed from fp */
+#ifndef _NO_POS_ARGS
+  int N;			/* arg number */
+  int arg_index = 0;		/* index into args processed directly */
+  int numargs = 0;		/* number of varargs read */
+  void *args[MAX_POS_ARGS];	/* positional args read */
+  int is_pos_arg;		/* is current format positional? */
+#endif
   int base = 0;			/* base argument to strtol/strtoul */
   int nbytes = 1;               /* number of bytes read from fmt string */
   wchar_t wc;                   /* wchar to use to read format string */
@@ -281,7 +455,8 @@ _DEFUN(__SVFSCANF_R, (rptr, fp, fmt0, ap),
   size_t mbslen;                /* length of converted multibyte sequence */
   mbstate_t state;              /* value to keep track of multibyte state */
 
-  u_long (*ccfn) () = 0;	/* conversion function (strtol/strtoul) */
+  #define CCFN_PARAMS	_PARAMS((struct _reent *, const char *, char **, int))
+  u_long (*ccfn)CCFN_PARAMS=0;	/* conversion function (strtol/strtoul) */
   char ccltab[256];		/* character class table for %[...] */
   char buf[BUF];		/* buffer for numeric conversions */
   char *lptr;                   /* literal pointer */
@@ -289,9 +464,11 @@ _DEFUN(__SVFSCANF_R, (rptr, fp, fmt0, ap),
   char *cp;
   short *sp;
   int *ip;
+#ifdef FLOATING_POINT
   float *flp;
   _LONG_DOUBLE *ldp;
   double *dp;
+#endif
   long *lp;
 #ifndef _NO_LONGLONG
   long long *llp;
@@ -301,8 +478,27 @@ _DEFUN(__SVFSCANF_R, (rptr, fp, fmt0, ap),
   static _CONST short basefix[17] =
     {10, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16};
 
+  /* Macro to support positional arguments */
+#ifndef _NO_POS_ARGS
+# define GET_ARG(n, ap, type)					\
+  ((type) (is_pos_arg						\
+	   ? (n < numargs					\
+	      ? args[n]						\
+	      : get_arg (n, &ap, &numargs, args))		\
+	   : (arg_index++ < numargs				\
+	      ? args[n]						\
+	      : (numargs < MAX_POS_ARGS				\
+		 ? args[numargs++] = va_arg (ap, void *)	\
+		 : va_arg (ap, void *)))))
+#else
+# define GET_ARG(n, ap, type) (va_arg (ap, type))
+#endif
+
+  __sfp_lock_acquire ();
   _flockfile (fp);
- 
+
+  ORIENT (fp, -1);
+
   nassigned = 0;
   nread = 0;
   for (;;)
@@ -311,7 +507,8 @@ _DEFUN(__SVFSCANF_R, (rptr, fp, fmt0, ap),
       wc = *fmt;
 #else
       memset (&state, '\0', sizeof (state));
-      nbytes = _mbtowc_r (rptr, &wc, fmt, MB_CUR_MAX, &state);
+      nbytes = __mbtowc (rptr, &wc, fmt, MB_CUR_MAX, __locale_charset (),
+			 &state);
 #endif
       fmt += nbytes;
       if (wc == 0)
@@ -330,6 +527,10 @@ _DEFUN(__SVFSCANF_R, (rptr, fp, fmt0, ap),
 	goto literal;
       width = 0;
       flags = 0;
+#ifndef _NO_POS_ARGS
+      N = arg_index;
+      is_pos_arg = 0;
+#endif
 
       /*
        * switch on the format.  continue if done; break once format
@@ -360,26 +561,69 @@ _DEFUN(__SVFSCANF_R, (rptr, fp, fmt0, ap),
 	  flags |= SUPPRESS;
 	  goto again;
 	case 'l':
+#if defined _WANT_IO_C99_FORMATS || !defined _NO_LONGLONG
 	  if (*fmt == 'l')	/* Check for 'll' = long long (SUSv3) */
 	    {
 	      ++fmt;
 	      flags |= LONGDBL;
 	    }
 	  else
+#endif
 	    flags |= LONG;
 	  goto again;
 	case 'L':
 	  flags |= LONGDBL;
 	  goto again;
 	case 'h':
+#ifdef _WANT_IO_C99_FORMATS
 	  if (*fmt == 'h')	/* Check for 'hh' = char int (SUSv3) */
 	    {
 	      ++fmt;
 	      flags |= CHAR;
 	    }
 	  else
+#endif
 	    flags |= SHORT;
 	  goto again;
+#ifdef _WANT_IO_C99_FORMATS
+	case 'j': /* intmax_t */
+	  if (sizeof (intmax_t) == sizeof (long))
+	    flags |= LONG;
+	  else
+	    flags |= LONGDBL;
+	  goto again;
+	case 't': /* ptrdiff_t */
+	  if (sizeof (ptrdiff_t) < sizeof (int))
+	    /* POSIX states ptrdiff_t is 16 or more bits, as
+	       is short.  */
+	    flags |= SHORT;
+	  else if (sizeof (ptrdiff_t) == sizeof (int))
+	    /* no flag needed */;
+	  else if (sizeof (ptrdiff_t) <= sizeof (long))
+	    flags |= LONG;
+	  else
+	    /* POSIX states that at least one programming
+	       environment must support ptrdiff_t no wider than
+	       long, but that means other environments can
+	       have ptrdiff_t as wide as long long.  */
+	    flags |= LONGDBL;
+	  goto again;
+	case 'z': /* size_t */
+	  if (sizeof (size_t) < sizeof (int))
+	    /* POSIX states size_t is 16 or more bits, as is short.  */
+	    flags |= SHORT;
+	  else if (sizeof (size_t) == sizeof (int))
+	    /* no flag needed */;
+	  else if (sizeof (size_t) <= sizeof (long))
+	    flags |= LONG;
+	  else
+	    /* POSIX states that at least one programming
+	       environment must support size_t no wider than
+	       long, but that means other environments can
+	       have size_t as wide as long long.  */
+	    flags |= LONGDBL;
+	  goto again;
+#endif /* _WANT_IO_C99_FORMATS */
 
 	case '0':
 	case '1':
@@ -394,6 +638,19 @@ _DEFUN(__SVFSCANF_R, (rptr, fp, fmt0, ap),
 	  width = width * 10 + c - '0';
 	  goto again;
 
+#ifndef _NO_POS_ARGS
+	case '$':
+	  if (width <= MAX_POS_ARGS)
+	    {
+	      N = width - 1;
+	      is_pos_arg = 1;
+	      width = 0;
+	      goto again;
+	    }
+	  rptr->_errno = EINVAL;
+	  goto input_failure;
+#endif /* !_NO_POS_ARGS */
+
 	  /*
 	   * Conversions. Those marked `compat' are for
 	   * 4.[123]BSD compatibility.
@@ -407,13 +664,13 @@ _DEFUN(__SVFSCANF_R, (rptr, fp, fmt0, ap),
 	  /* FALLTHROUGH */
 	case 'd':
 	  c = CT_INT;
-	  ccfn = (u_long (*)())_strtol_r;
+	  ccfn = (u_long (*)CCFN_PARAMS)_strtol_r;
 	  base = 10;
 	  break;
 
 	case 'i':
 	  c = CT_INT;
-	  ccfn = (u_long (*)())_strtol_r;
+	  ccfn = (u_long (*)CCFN_PARAMS)_strtol_r;
 	  base = 0;
 	  break;
 
@@ -432,7 +689,7 @@ _DEFUN(__SVFSCANF_R, (rptr, fp, fmt0, ap),
 	  base = 10;
 	  break;
 
-	case 'X':		/* compat   XXX */
+	case 'X':
 	case 'x':
 	  flags |= PFXOK;	/* enable 0x prefixing */
 	  c = CT_INT;
@@ -441,33 +698,41 @@ _DEFUN(__SVFSCANF_R, (rptr, fp, fmt0, ap),
 	  break;
 
 #ifdef FLOATING_POINT
-	case 'E':		/* compat   XXX */
-	case 'G':		/* compat   XXX */
-/* ANSI says that E,G and X behave the same way as e,g,x */
-	  /* FALLTHROUGH */
+# ifdef _WANT_IO_C99_FORMATS
+	case 'a':
+	case 'A':
+	case 'F':
+# endif
+	case 'E':
+	case 'G':
 	case 'e':
 	case 'f':
 	case 'g':
 	  c = CT_FLOAT;
 	  break;
 #endif
-        case 'S':
-          flags |= LONG;
-          /* FALLTHROUGH */
+
+#ifdef _WANT_IO_C99_FORMATS
+	case 'S':
+	  flags |= LONG;
+	  /* FALLTHROUGH */
+#endif
 
 	case 's':
 	  c = CT_STRING;
 	  break;
 
 	case '[':
-	  fmt = __sccl (ccltab, fmt);
+	  fmt = (u_char *) __sccl (ccltab, (unsigned char *) fmt);
 	  flags |= NOSKIP;
 	  c = CT_CCL;
 	  break;
 
-        case 'C':
-          flags |= LONG;
-          /* FALLTHROUGH */
+#ifdef _WANT_IO_C99_FORMATS
+	case 'C':
+	  flags |= LONG;
+	  /* FALLTHROUGH */
+#endif
 
 	case 'c':
 	  flags |= NOSKIP;
@@ -484,31 +749,34 @@ _DEFUN(__SVFSCANF_R, (rptr, fp, fmt0, ap),
 	case 'n':
 	  if (flags & SUPPRESS)	/* ??? */
 	    continue;
+#ifdef _WANT_IO_C99_FORMATS
 	  if (flags & CHAR)
 	    {
-	      cp = va_arg (ap, char *);
+	      cp = GET_ARG (N, ap, char *);
 	      *cp = nread;
 	    }
-	  else if (flags & SHORT)
+	  else
+#endif
+	  if (flags & SHORT)
 	    {
-	      sp = va_arg (ap, short *);
+	      sp = GET_ARG (N, ap, short *);
 	      *sp = nread;
 	    }
 	  else if (flags & LONG)
 	    {
-	      lp = va_arg (ap, long *);
+	      lp = GET_ARG (N, ap, long *);
 	      *lp = nread;
 	    }
 #ifndef _NO_LONGLONG
 	  else if (flags & LONGDBL)
 	    {
-	      llp = va_arg (ap, long long*);
+	      llp = GET_ARG (N, ap, long long*);
 	      *llp = nread;
 	    }
 #endif
 	  else
 	    {
-	      ip = va_arg (ap, int *);
+	      ip = GET_ARG (N, ap, int *);
 	      *ip = nread;
 	    }
 	  continue;
@@ -518,13 +786,14 @@ _DEFUN(__SVFSCANF_R, (rptr, fp, fmt0, ap),
 	   */
 	case '\0':		/* compat */
 	  _funlockfile (fp);
+	  __sfp_lock_release ();
 	  return EOF;
 
 	default:		/* compat */
 	  if (isupper (c))
 	    flags |= LONG;
 	  c = CT_INT;
-	  ccfn = (u_long (*)())_strtol_r;
+	  ccfn = (u_long (*)CCFN_PARAMS)_strtol_r;
 	  base = 10;
 	  break;
 	}
@@ -567,14 +836,15 @@ _DEFUN(__SVFSCANF_R, (rptr, fp, fmt0, ap),
 	  /* scan arbitrary characters (sets NOSKIP) */
 	  if (width == 0)
 	    width = 1;
-          if (flags & LONG) 
+#if !defined(_ELIX_LEVEL) || _ELIX_LEVEL >= 2
+          if (flags & LONG)
             {
               if ((flags & SUPPRESS) == 0)
-                wcp = va_arg (ap, wchar_t *);
+                wcp = GET_ARG (N, ap, wchar_t *);
               else
                 wcp = NULL;
               n = 0;
-              while (width != 0) 
+              while (width != 0)
                 {
                   if (n == MB_CUR_MAX)
                     goto input_failure;
@@ -582,7 +852,7 @@ _DEFUN(__SVFSCANF_R, (rptr, fp, fmt0, ap),
                   fp->_r -= 1;
                   fp->_p += 1;
                   memset ((_PTR)&state, '\0', sizeof (mbstate_t));
-                  if ((mbslen = _mbrtowc_r (rptr, wcp, buf, n, &state)) 
+                  if ((mbslen = _mbrtowc_r (rptr, wcp, buf, n, &state))
                                                          == (size_t)-1)
                     goto input_failure; /* Invalid sequence */
                   if (mbslen == 0 && !(flags & SUPPRESS))
@@ -595,17 +865,19 @@ _DEFUN(__SVFSCANF_R, (rptr, fp, fmt0, ap),
                         wcp += 1;
                       n = 0;
                     }
-                  if (BufferEmpty) 
+                  if (BufferEmpty)
 	            {
-                      if (n != 0) 
+                      if (n != 0)
                         goto input_failure;
                       break;
                     }
                 }
               if (!(flags & SUPPRESS))
                 nassigned++;
-            } 
-          else if (flags & SUPPRESS) 
+            }
+          else
+#endif
+        	  if (flags & SUPPRESS)
 	    {
 	      size_t sum = 0;
 	      for (;;)
@@ -634,7 +906,7 @@ _DEFUN(__SVFSCANF_R, (rptr, fp, fmt0, ap),
 	    }
 	  else
 	    {
-	      size_t r = fread ((_PTR) va_arg (ap, char *), 1, width, fp);
+	      size_t r = _fread_r (rptr, (_PTR) GET_ARG (N, ap, char *), 1, width, fp);
 
 	      if (r == 0)
 		goto input_failure;
@@ -668,7 +940,7 @@ _DEFUN(__SVFSCANF_R, (rptr, fp, fmt0, ap),
 	    }
 	  else
 	    {
-	      p0 = p = va_arg (ap, char *);
+	      p0 = p = GET_ARG (N, ap, char *);
 	      while (ccltab[*fp->_p])
 		{
 		  fp->_r--;
@@ -695,15 +967,16 @@ _DEFUN(__SVFSCANF_R, (rptr, fp, fmt0, ap),
 	  /* like CCL, but zero-length string OK, & no NOSKIP */
 	  if (width == 0)
             width = (size_t)~0;
-          if (flags & LONG) 
+#if !defined(_ELIX_LEVEL) || _ELIX_LEVEL >= 2
+          if (flags & LONG)
             {
               /* Process %S and %ls placeholders */
               if ((flags & SUPPRESS) == 0)
-                wcp = va_arg (ap, wchar_t *);
+                wcp = GET_ARG (N, ap, wchar_t *);
               else
                 wcp = &wc;
               n = 0;
-              while (!isspace (*fp->_p) && width != 0) 
+              while (!isspace (*fp->_p) && width != 0)
                 {
                   if (n == MB_CUR_MAX)
                     goto input_failure;
@@ -711,17 +984,17 @@ _DEFUN(__SVFSCANF_R, (rptr, fp, fmt0, ap),
                   fp->_r -= 1;
                   fp->_p += 1;
                   memset ((_PTR)&state, '\0', sizeof (mbstate_t));
-                  if ((mbslen = _mbrtowc_r (rptr, wcp, buf, n, &state)) 
+                  if ((mbslen = _mbrtowc_r (rptr, wcp, buf, n, &state))
                                                         == (size_t)-1)
                     goto input_failure;
                   if (mbslen == 0)
                     *wcp = L'\0';
                   if (mbslen != (size_t)-2) /* Incomplete sequence */
                     {
-                      if (iswspace(*wcp)) 
+                      if (iswspace(*wcp))
                         {
                           while (n != 0)
-                            ungetc (buf[--n], fp);
+                            _ungetc_r (rptr, (unsigned char) buf[--n], fp);
                           break;
                         }
                       nread += n;
@@ -730,20 +1003,22 @@ _DEFUN(__SVFSCANF_R, (rptr, fp, fmt0, ap),
                         wcp += 1;
                       n = 0;
                     }
-                  if (BufferEmpty) 
+                  if (BufferEmpty)
                     {
                       if (n != 0)
                         goto input_failure;
                       break;
                     }
                 }
-              if (!(flags & SUPPRESS)) 
+              if (!(flags & SUPPRESS))
                 {
                   *wcp = L'\0';
                   nassigned++;
                 }
             }
-          else if (flags & SUPPRESS) 
+          else
+#endif
+        	  if (flags & SUPPRESS)
 	    {
 	      n = 0;
 	      while (!isspace (*fp->_p))
@@ -758,7 +1033,7 @@ _DEFUN(__SVFSCANF_R, (rptr, fp, fmt0, ap),
 	    }
 	  else
 	    {
-	      p0 = p = va_arg (ap, char *);
+	      p0 = p = GET_ARG (N, ap, char *);
 	      while (!isspace (*fp->_p))
 		{
 		  fp->_r--;
@@ -880,15 +1155,15 @@ _DEFUN(__SVFSCANF_R, (rptr, fp, fmt0, ap),
 		    }
 		  break;
 
-		  /* x ok iff flag still set & 2nd char */
+		  /* x ok iff flag still set & single 0 seen */
 		case 'x':
 		case 'X':
-		  if (flags & PFXOK && p == buf + 1)
+		  if ((flags & (PFXOK | NZDIGITS)) == PFXOK)
 		    {
 		      base = 16;/* if %i */
 		      flags &= ~PFXOK;
 		      /* We must reset the NZDIGITS and NDIGITS
-		         flags that would have been unset by seeing
+			 flags that would have been unset by seeing
 			 the zero that preceded the X or x.  */
 		      flags |= NZDIGITS | NDIGITS;
 		      goto ok;
@@ -917,18 +1192,16 @@ _DEFUN(__SVFSCANF_R, (rptr, fp, fmt0, ap),
 	   * If we had only a sign, it is no good; push back the sign.
 	   * If the number ends in `x', it was [sign] '0' 'x', so push back
 	   * the x and treat it as [sign] '0'.
+	   * Use of ungetc here and below assumes ASCII encoding; we are only
+	   * pushing back 7-bit characters, so casting to unsigned char is
+	   * not necessary.
 	   */
 	  if (flags & NDIGITS)
 	    {
 	      if (p > buf)
-		_CAST_VOID ungetc (*(u_char *)-- p, fp);
-	      goto match_failure;
-	    }
-	  c = ((u_char *) p)[-1];
-	  if (c == 'x' || c == 'X')
-	    {
-	      --p;
-	      /*(void)*/ ungetc (c, fp);
+		_ungetc_r (rptr, *--p, fp); /* [-+xX] */
+	      if (p == buf)
+		goto match_failure;
 	    }
 	  if ((flags & SUPPRESS) == 0)
 	    {
@@ -937,20 +1210,34 @@ _DEFUN(__SVFSCANF_R, (rptr, fp, fmt0, ap),
 	      *p = 0;
 	      res = (*ccfn) (rptr, buf, (char **) NULL, base);
 	      if (flags & POINTER)
-		*(va_arg (ap, _PTR *)) = (_PTR) (unsigned _POINTER_INT) res;
+		{
+		  void **vp = GET_ARG (N, ap, void **);
+#ifndef _NO_LONGLONG
+		  if (sizeof (uintptr_t) > sizeof (u_long))
+		    {
+		      u_long_long resll;
+		      resll = _strtoull_r (rptr, buf, (char **) NULL, base);
+		      *vp = (void *) (uintptr_t) resll;
+		    }
+		  else
+#endif /* !_NO_LONGLONG */
+		    *vp = (void *) (uintptr_t) res;
+		}
+#ifdef _WANT_IO_C99_FORMATS
 	      else if (flags & CHAR)
 		{
-		  cp = va_arg (ap, char *);
+		  cp = GET_ARG (N, ap, char *);
 		  *cp = res;
 		}
+#endif
 	      else if (flags & SHORT)
 		{
-		  sp = va_arg (ap, short *);
+		  sp = GET_ARG (N, ap, short *);
 		  *sp = res;
 		}
 	      else if (flags & LONG)
 		{
-		  lp = va_arg (ap, long *);
+		  lp = GET_ARG (N, ap, long *);
 		  *lp = res;
 		}
 #ifndef _NO_LONGLONG
@@ -961,13 +1248,13 @@ _DEFUN(__SVFSCANF_R, (rptr, fp, fmt0, ap),
 		    resll = _strtoull_r (rptr, buf, (char **) NULL, base);
 		  else
 		    resll = _strtoll_r (rptr, buf, (char **) NULL, base);
-		  llp = va_arg (ap, long long*);
+		  llp = GET_ARG (N, ap, long long*);
 		  *llp = resll;
 		}
 #endif
 	      else
 		{
-		  ip = va_arg (ap, int *);
+		  ip = GET_ARG (N, ap, int *);
 		  *ip = res;
 		}
 	      nassigned++;
@@ -987,7 +1274,8 @@ _DEFUN(__SVFSCANF_R, (rptr, fp, fmt0, ap),
 	  long zeroes, exp_adjust;
 	  char *exp_start = NULL;
 	  unsigned width_left = 0;
-	  int nancount = 0;
+	  char nancount = 0;
+	  char infcount = 0;
 #ifdef hardway
 	  if (width == 0 || width > sizeof (buf) - 1)
 #else
@@ -1032,7 +1320,7 @@ _DEFUN(__SVFSCANF_R, (rptr, fp, fmt0, ap),
 		case '7':
 		case '8':
 		case '9':
-		  if (nancount == 0)
+		  if (nancount + infcount == 0)
 		    {
 		      flags &= ~(SIGNOK | NDIGITS);
 		      goto fok;
@@ -1049,19 +1337,24 @@ _DEFUN(__SVFSCANF_R, (rptr, fp, fmt0, ap),
 		  break;
 		case 'n':
 		case 'N':
-	          if (nancount == 0
-		      && (flags & (SIGNOK | NDIGITS | DPTOK | EXPOK)) ==
-		      	          (SIGNOK | NDIGITS | DPTOK | EXPOK))
+		  if (nancount == 0 && zeroes == 0
+		      && (flags & (NDIGITS | DPTOK | EXPOK)) ==
+				  (NDIGITS | DPTOK | EXPOK))
 		    {
 		      flags &= ~(SIGNOK | DPTOK | EXPOK | NDIGITS);
 		      nancount = 1;
 		      goto fok;
 		    }
-		  else if (nancount == 2)
+		  if (nancount == 2)
 		    {
 		      nancount = 3;
 		      goto fok;
 		    }
+		  if (infcount == 1 || infcount == 4)
+		    {
+		      infcount++;
+		      goto fok;
+		    }
 		  break;
 		case 'a':
 		case 'A':
@@ -1071,6 +1364,46 @@ _DEFUN(__SVFSCANF_R, (rptr, fp, fmt0, ap),
 		      goto fok;
 		    }
 		  break;
+		case 'i':
+		case 'I':
+		  if (infcount == 0 && zeroes == 0
+		      && (flags & (NDIGITS | DPTOK | EXPOK)) ==
+				  (NDIGITS | DPTOK | EXPOK))
+		    {
+		      flags &= ~(SIGNOK | DPTOK | EXPOK | NDIGITS);
+		      infcount = 1;
+		      goto fok;
+		    }
+		  if (infcount == 3 || infcount == 5)
+		    {
+		      infcount++;
+		      goto fok;
+		    }
+		  break;
+		case 'f':
+		case 'F':
+		  if (infcount == 2)
+		    {
+		      infcount = 3;
+		      goto fok;
+		    }
+		  break;
+		case 't':
+		case 'T':
+		  if (infcount == 6)
+		    {
+		      infcount = 7;
+		      goto fok;
+		    }
+		  break;
+		case 'y':
+		case 'Y':
+		  if (infcount == 7)
+		    {
+		      infcount = 8;
+		      goto fok;
+		    }
+		  break;
 		case '.':
 		  if (flags & DPTOK)
 		    {
@@ -1103,7 +1436,7 @@ _DEFUN(__SVFSCANF_R, (rptr, fp, fmt0, ap),
 	      *p++ = c;
 	    fskip:
 	      width--;
-              ++nread;
+	      ++nread;
 	      if (--fp->_r > 0)
 		fp->_p++;
 	      else
@@ -1112,24 +1445,48 @@ _DEFUN(__SVFSCANF_R, (rptr, fp, fmt0, ap),
 	    }
 	  if (zeroes)
 	    flags &= ~NDIGITS;
-          /* We may have a 'N' or possibly even a 'Na' as the start of 'NaN', 
-	     only to run out of chars before it was complete (or having 
-	     encountered a non- matching char).  So check here if we have an 
-	     outstanding nancount, and if so put back the chars we did 
-	     swallow and treat as a failed match. */
-          if (nancount && nancount != 3)
-            {
-              /* Ok... what are we supposed to do in the event that the
-              __srefill call above was triggered in the middle of the partial
-              'NaN' and so we can't put it all back? */
-              while (nancount-- && (p > buf))
-                {
-                  ungetc (*(u_char *)--p, fp);
-                  --nread;
-                }
-              goto match_failure;
-            }
-          /*
+	  /* We may have a 'N' or possibly even [sign] 'N' 'a' as the
+	     start of 'NaN', only to run out of chars before it was
+	     complete (or having encountered a non-matching char).  So
+	     check here if we have an outstanding nancount, and if so
+	     put back the chars we did swallow and treat as a failed
+	     match.
+
+	     FIXME - we still don't handle NAN([0xdigits]).  */
+	  if (nancount - 1U < 2U) /* nancount && nancount < 3 */
+	    {
+	      /* Newlib's ungetc works even if we called __srefill in
+		 the middle of a partial parse, but POSIX does not
+		 guarantee that in all implementations of ungetc.  */
+	      while (p > buf)
+		{
+		  _ungetc_r (rptr, *--p, fp); /* [-+nNaA] */
+		  --nread;
+		}
+	      goto match_failure;
+	    }
+	  /* Likewise for 'inf' and 'infinity'.	 But be careful that
+	     'infinite' consumes only 3 characters, leaving the stream
+	     at the second 'i'.	 */
+	  if (infcount - 1U < 7U) /* infcount && infcount < 8 */
+	    {
+	      if (infcount >= 3) /* valid 'inf', but short of 'infinity' */
+		while (infcount-- > 3)
+		  {
+		    _ungetc_r (rptr, *--p, fp); /* [iInNtT] */
+		    --nread;
+		  }
+	      else
+		{
+		  while (p > buf)
+		    {
+		      _ungetc_r (rptr, *--p, fp); /* [-+iInN] */
+		      --nread;
+		    }
+		  goto match_failure;
+		}
+	    }
+	  /*
 	   * If no digits, might be missing exponent digits
 	   * (just give back the exponent) or might be missing
 	   * regular digits, but had sign and/or decimal point.
@@ -1140,22 +1497,22 @@ _DEFUN(__SVFSCANF_R, (rptr, fp, fmt0, ap),
 		{
 		  /* no digits at all */
 		  while (p > buf)
-                    {
-		      ungetc (*(u_char *)--p, fp);
-                      --nread;
-                    }
+		    {
+		      _ungetc_r (rptr, *--p, fp); /* [-+.] */
+		      --nread;
+		    }
 		  goto match_failure;
 		}
 	      /* just a bad exponent (e and maybe sign) */
-	      c = *(u_char *)-- p;
-              --nread;
+	      c = *--p;
+	      --nread;
 	      if (c != 'e' && c != 'E')
 		{
-		  _CAST_VOID ungetc (c, fp);	/* sign */
-		  c = *(u_char *)-- p;
-                  --nread;
+		  _ungetc_r (rptr, c, fp); /* [-+] */
+		  c = *--p;
+		  --nread;
 		}
-	      _CAST_VOID ungetc (c, fp);
+	      _ungetc_r (rptr, c, fp); /* [eE] */
 	    }
 	  if ((flags & SUPPRESS) == 0)
 	    {
@@ -1187,29 +1544,29 @@ _DEFUN(__SVFSCANF_R, (rptr, fp, fmt0, ap),
                  sprintf (exp_start, "e%ld", new_exp);
 		}
 
-	      /* Current _strtold routine is markedly slower than 
+	      /* Current _strtold routine is markedly slower than
 	         _strtod_r.  Only use it if we have a long double
 	         result.  */
 #ifndef _NO_LONGDBL /* !_NO_LONGDBL */
 	      if (flags & LONGDBL)
-	      	qres = _strtold (buf, NULL);
+		qres = _strtold (buf, NULL);
 	      else
 #endif
 	        res = _strtod_r (rptr, buf, NULL);
 
 	      if (flags & LONG)
 		{
-		  dp = va_arg (ap, double *);
+		  dp = GET_ARG (N, ap, double *);
 		  *dp = res;
 		}
 	      else if (flags & LONGDBL)
 		{
-		  ldp = va_arg (ap, _LONG_DOUBLE *);
+		  ldp = GET_ARG (N, ap, _LONG_DOUBLE *);
 		  *ldp = QUAD_RES;
 		}
 	      else
 		{
-		  flp = va_arg (ap, float *);
+		  flp = GET_ARG (N, ap, float *);
 		  if (isnan (res))
 		    *flp = nanf (NULL);
 		  else
@@ -1223,10 +1580,32 @@ _DEFUN(__SVFSCANF_R, (rptr, fp, fmt0, ap),
 	}
     }
 input_failure:
+  /* On read failure, return EOF failure regardless of matches; errno
+     should have been set prior to here.  On EOF failure (including
+     invalid format string), return EOF if no matches yet, else number
+     of matches made prior to failure.  */
   _funlockfile (fp);
-  return nassigned ? nassigned : -1;
+  __sfp_lock_release ();
+  return nassigned && !(fp->_flags & __SERR) ? nassigned : EOF;
 match_failure:
 all_done:
+  /* Return number of matches, which can be 0 on match failure.  */
   _funlockfile (fp);
+  __sfp_lock_release ();
   return nassigned;
 }
+
+#ifndef _NO_POS_ARGS
+/* Process all intermediate arguments.  Fortunately, with scanf, all
+   intermediate arguments are sizeof(void*), so we don't need to scan
+   ahead in the format string.  */
+static void *
+get_arg (int n, va_list *ap, int *numargs_p, void **args)
+{
+  int numargs = *numargs_p;
+  while (n >= numargs)
+    args[numargs++] = va_arg (*ap, void *);
+  *numargs_p = numargs;
+  return args[n];
+}
+#endif /* !_NO_POS_ARGS */
