@@ -7,6 +7,9 @@
 #include <reent.h>
 #include "atexit.h"
 
+/* Make this a weak reference to avoid pulling in free.  */
+void free(void *) _ATTRIBUTE((__weak__));
+
 /*
  * Call registered exit handlers.  If D is null then all handlers are called,
  * otherwise only the handlers from that DSO are called.
@@ -23,6 +26,8 @@ _DEFUN (__call_exitprocs, (code, d),
   int i;
   void (*fn) (void);
 
+ restart:
+
   p = _GLOBAL_REENT->_atexit;
   lastp = &_GLOBAL_REENT->_atexit;
   while (p)
@@ -34,6 +39,8 @@ _DEFUN (__call_exitprocs, (code, d),
 #endif
       for (n = p->_ind - 1; n >= 0; n--)
 	{
+	  int ind;
+
 	  i = 1 << n;
 
 	  /* Skip functions not from this dso.  */
@@ -52,6 +59,8 @@ _DEFUN (__call_exitprocs, (code, d),
 	  if (!fn)
 	    continue;
 
+	  ind = p->_ind;
+
 	  /* Call the function.  */
 	  if (!args || (args->_fntypes & i) == 0)
 	    fn ();
@@ -59,11 +68,22 @@ _DEFUN (__call_exitprocs, (code, d),
 	    (*((void (*)(int, _PTR)) fn))(code, args->_fnargs[n]);
 	  else
 	    (*((void (*)(_PTR)) fn))(args->_fnargs[n]);
+
+	  /* The function we called call atexit and registered another
+	     function (or functions).  Call these new functions before
+	     continuing with the already registered functions.  */
+	  if (ind != p->_ind || *lastp != p)
+	    goto restart;
 	}
 
 #ifndef _ATEXIT_DYNAMIC_ALLOC
       break;
 #else
+      /* Don't dynamically free the atexit array if free is not
+	 available.  */
+      if (!free)
+	break;
+
       /* Move to the next block.  Free empty blocks except the last one,
 	 which is part of _GLOBAL_REENT.  */
       if (p->_ind == 0 && p->_next)
