@@ -29,14 +29,14 @@ A pointer to the (string) value of the environment variable, or
 
 PORTABILITY
 <<_getenv_r>> is not ANSI; the rules for properly forming names of environment
-variables vary from one system to another.
+variables vary from one system to another.  This implementation does not
+permit '=' to be in identifiers.
 
 <<_getenv_r>> requires a global pointer <<environ>>.
 */
 
 /* This file may have been modified by DJ Delorie (Jan 1991).  If so,
-** these modifications are Coyright (C) 1991 DJ Delorie, 24 Kirsten Ave,
-** Rochester NH, 03867-2954, USA.
+** these modifications are Copyright (C) 1991 DJ Delorie.
 */
 
 /*
@@ -75,7 +75,6 @@ static char ***p_environ = &environ;
  *	Returns pointer to value associated with name, if any, else NULL.
  *	Sets offset to be the offset of the name/value combination in the
  *	environmental array, for use by setenv(3) and unsetenv(3).
- *	Explicitly removes '=' in argument name.
  *
  *	This routine *should* be a static; don't use it.
  */
@@ -95,24 +94,27 @@ _DEFUN (_findenv_r, (reent_ptr, name, offset),
   /* In some embedded systems, this does not get set.  This protects
      newlib from dereferencing a bad pointer.  */
   if (!*p_environ)
-    return NULL;
-
-  c = name;
-  len = 0;
-  while (*c && *c != '=')
     {
-      c++;
-      len++;
+      ENV_UNLOCK;
+      return NULL;
     }
 
-  for (p = *p_environ; *p; ++p)
-    if (!strncmp (*p, name, len))
-      if (*(c = *p + len) == '=')
+  c = name;
+  while (*c && *c != '=')  c++;
+ 
+  /* Identifiers may not contain an '=', so cannot match if does */
+  if(*c != '=')
+    {
+    len = c - name;
+    for (p = *p_environ; *p; ++p)
+      if (!strncmp (*p, name, len))
+        if (*(c = *p + len) == '=')
 	{
 	  *offset = p - *p_environ;
-          ENV_UNLOCK;
+	  ENV_UNLOCK;
 	  return (char *) (++c);
 	}
+    }
   ENV_UNLOCK;
   return NULL;
 }
