@@ -78,31 +78,18 @@ union double_union
 #define word1(x) (x.i[1])
 #endif
 
-
-/* The following is taken from gdtoaimp.h for use with new strtod.  */
+/* The following is taken from gdtoaimp.h for use with new strtod, but
+   adjusted to avoid invalid type-punning.  */
 typedef __int32_t Long;
-typedef union { double d; __ULong L[2]; } U;
 
-#ifdef YES_ALIAS
-#define dval(x) x
-#ifdef IEEE_8087
-#define dword0(x) ((__ULong *)&x)[1]
-#define dword1(x) ((__ULong *)&x)[0]
-#else
-#define dword0(x) ((__ULong *)&x)[0]
-#define dword1(x) ((__ULong *)&x)[1]
-#endif
-#else /* !YES_ALIAS */
-#ifdef IEEE_8087
-#define dword0(x) ((U*)&x)->L[1]
-#define dword1(x) ((U*)&x)->L[0]
-#else
-#define dword0(x) ((U*)&x)->L[0]
-#define dword1(x) ((U*)&x)->L[1]
-#endif
-#define dval(x) ((U*)&x)->d
-#endif /* YES_ALIAS */
+/* Unfortunately, because __ULong might be a different type than
+   __uint32_t, we can't re-use union double_union as-is without
+   further edits in strtod.c.  */
+typedef union { double d; __ULong i[2]; } U;
 
+#define dword0(x) word0(x)
+#define dword1(x) word1(x)
+#define dval(x) (x.d)
 
 #undef SI
 #ifdef Sudden_Underflow
@@ -111,17 +98,7 @@ typedef union { double d; __ULong L[2]; } U;
 #define SI 0
 #endif
 
-/* The following definition of Storeinc is appropriate for MIPS processors.
- * An alternative that might be better on some machines is
- * #define Storeinc(a,b,c) (*a++ = b << 16 | c & 0xffff)
- */
-#if defined (__IEEE_BYTES_LITTLE_ENDIAN) + defined (IEEE_8087) + defined (VAX)
-#define Storeinc(a,b,c) (((unsigned short *)a)[1] = (unsigned short)b, \
-((unsigned short *)a)[0] = (unsigned short)c, a++)
-#else
-#define Storeinc(a,b,c) (((unsigned short *)a)[0] = (unsigned short)b, \
-((unsigned short *)a)[1] = (unsigned short)c, a++)
-#endif
+#define Storeinc(a,b,c) (*(a)++ = (b) << 16 | (c) & 0xffff)
 
 /* #define P DBL_MANT_DIG */
 /* Ten_pmax = floor(P*log(2)/log(5)) */
@@ -139,15 +116,7 @@ typedef union { double d; __ULong L[2]; } U;
 #define P    	    24
 #define Bias 	    127
 #define NO_HEX_FP   /* not supported in this case */
-#if 0
-#define IEEE_Arith  /* it is, but the code doesn't handle IEEE singles yet */
-#endif
-/* Following is needed due to IEEE_Arith not being set on above.  */
-#if defined(__v800)
-#define n_bigtens 2
-#else
-#define n_bigtens 5
-#endif
+#define IEEE_Arith
 #define Emin        (-126)
 #define Exp_1       ((__uint32_t)0x3f800000L)
 #define Exp_11      ((__uint32_t)0x3f800000L)
@@ -175,7 +144,7 @@ typedef union { double d; __ULong L[2]; } U;
 
 #define word0(x) (x.i[0])
 #define word1(x) 0
-#define dword0(x) ((__ULong *)&x)[0]
+#define dword0(x) word0(x)
 #define dword1(x) 0
 #else
 
@@ -351,10 +320,14 @@ extern double rnd_prod(double, double), rnd_quot(double, double);
  * slower.  Hence the default is now to store 32 bits per long.
  */
 
-#ifndef Pack_32
-#define Pack_32
-#endif
-#endif
+ #ifndef Pack_32
+  #define Pack_32
+ #endif
+#else  /* Just_16 */
+ #ifndef Pack_16
+  #define Pack_16
+ #endif
+#endif /* Just_16 */
 
 #ifdef Pack_32
 #define ULbits 32
