@@ -1,6 +1,5 @@
 /* This file may have been modified by DJ Delorie (Jan 1991).  If so,
-** these modifications are Coyright (C) 1991 DJ Delorie, 24 Kirsten Ave,
-** Rochester NH, 03867-2954, USA.
+** these modifications are Copyright (C) 1991 DJ Delorie.
 */
 
 /*
@@ -27,6 +26,8 @@
 #include <stddef.h>
 #include <stdlib.h>
 #include <string.h>
+#include <time.h>
+#include <errno.h>
 #include "envlock.h"
 
 extern char **environ;
@@ -43,6 +44,8 @@ extern char *_findenv_r _PARAMS ((struct _reent *, const char *, int *));
  * _setenv_r --
  *	Set the value of the environmental variable "name" to be
  *	"value".  If rewrite is set, replace any current value.
+ *	If "name" contains equal sign, -1 is returned, and errno is
+ *	set to EINVAL;
  */
 
 int
@@ -56,10 +59,14 @@ _DEFUN (_setenv_r, (reent_ptr, name, value, rewrite),
   register char *C;
   int l_value, offset;
 
+  if (strchr(name, '='))
+    {
+      errno = EINVAL;
+      return -1;
+    }
+
   ENV_LOCK;
 
-  if (*value == '=')		/* no `=' in value */
-    ++value;
   l_value = strlen (value);
   if ((C = _findenv_r (reent_ptr, name, &offset)))
     {				/* find if already exists */
@@ -70,8 +77,11 @@ _DEFUN (_setenv_r, (reent_ptr, name, value, rewrite),
         }
       if (strlen (C) >= l_value)
 	{			/* old larger; copy over */
-         while ((*C++ = *value++) != 0);
+	  while ((*C++ = *value++) != 0);
           ENV_UNLOCK;
+	  /* if we are changing the TZ environment variable, update timezone info */
+	  if (strcmp (name, "TZ") == 0)
+	    tzset ();
 	  return 0;
 	}
     }
@@ -118,6 +128,10 @@ _DEFUN (_setenv_r, (reent_ptr, name, value, rewrite),
 
   ENV_UNLOCK;
 
+  /* if we are setting the TZ environment variable, update timezone info */
+  if (strncmp ((*p_environ)[offset], "TZ=", 3) == 0)
+    tzset ();
+
   return 0;
 }
 
@@ -125,20 +139,30 @@ _DEFUN (_setenv_r, (reent_ptr, name, value, rewrite),
  * _unsetenv_r(name) --
  *	Delete environmental variable "name".
  */
-void
+int
 _DEFUN (_unsetenv_r, (reent_ptr, name),
         struct _reent *reent_ptr _AND
         _CONST char *name)
 {
   register char **P;
   int offset;
+ 
+  /* Name cannot be NULL, empty, or contain an equal sign.  */ 
+  if (name == NULL || name[0] == '\0' || strchr(name, '='))
+    {
+      errno = EINVAL;
+      return -1;
+    }
 
   ENV_LOCK;
 
   while (_findenv_r (reent_ptr, name, &offset))	/* if set multiple times */
-    for (P = &(*p_environ)[offset];; ++P)
-      if (!(*P = *(P + 1)))
-	break;
+    { 
+      for (P = &(*p_environ)[offset];; ++P)
+        if (!(*P = *(P + 1)))
+	  break;
+    }
 
   ENV_UNLOCK;
+  return 0;
 }
