@@ -12,17 +12,19 @@
 #include <sys/times.h>
 #include <errno.h>
 #include <reent.h>
+#include <signal.h>
 #include <unistd.h>
+#include <sys/wait.h>
 #include "swi.h"
 
 /* Forward prototypes.  */
 int     _system     _PARAMS ((const char *));
 int     _rename     _PARAMS ((const char *, const char *));
-int     isatty		_PARAMS ((int));
+int     _isatty		_PARAMS ((int));
 clock_t _times		_PARAMS ((struct tms *));
-int     _gettimeofday	_PARAMS ((struct timeval *, struct timezone *));
+int     _gettimeofday	_PARAMS ((struct timeval *, void *));
 void    _raise 		_PARAMS ((void));
-int     _unlink		_PARAMS ((void));
+int     _unlink		_PARAMS ((const char *));
 int     _link 		_PARAMS ((void));
 int     _stat 		_PARAMS ((const char *, struct stat *));
 int     _fstat 		_PARAMS ((int, struct stat *));
@@ -98,7 +100,7 @@ static inline int
 do_AngelSWI (int reason, void * arg)
 {
   int value;
-  asm volatile ("mov r0, %1; mov r1, %2; swi %a3; mov %0, r0"
+  asm volatile ("mov r0, %1; mov r1, %2; " AngelSWIInsn " %a3; mov %0, r0"
        : "=r" (value) /* Outputs */
        : "r" (reason), "r" (arg), "i" (AngelSWI) /* Inputs */
        : "r0", "r1", "r2", "r3", "ip", "lr", "memory", "cc"
@@ -433,30 +435,34 @@ _close (int file)
   return wrap (_swiclose (file));
 }
 
-void
-_exit (int n)
+int
+_kill (int pid, int sig)
 {
-  /* FIXME: return code is thrown away.  */
-  
+  (void)pid; (void)sig;
 #ifdef ARM_RDI_MONITOR
-  do_AngelSWI (AngelSWI_Reason_ReportException,
-	      (void *) ADP_Stopped_ApplicationExit);
+  /* Note: The pid argument is thrown away.  */
+  switch (sig) {
+	  case SIGABRT:
+		  return do_AngelSWI (AngelSWI_Reason_ReportException,
+				  (void *) ADP_Stopped_RunTimeError);
+	  default:
+		  return do_AngelSWI (AngelSWI_Reason_ReportException,
+				  (void *) ADP_Stopped_ApplicationExit);
+  }
 #else
   asm ("swi %a0" :: "i" (SWI_Exit));
 #endif
-  n = n;
 }
 
-int
-_kill (int n, int m)
+void
+_exit (int status)
 {
-#ifdef ARM_RDI_MONITOR
-  return do_AngelSWI (AngelSWI_Reason_ReportException,
-		      (void *) ADP_Stopped_ApplicationExit);
-#else
-  asm ("swi %a0" :: "i" (SWI_Exit));
-#endif
-  n = n; m = m;
+  /* There is only one SWI for both _exit and _kill. For _exit, call
+     the SWI with the second argument set to -1, an invalid value for
+     signum, so that the SWI handler can distinguish the two calls.
+     Note: The RDI implementation of _kill throws away both its
+     arguments.  */
+  _kill(status, -1);
 }
 
 int
@@ -534,9 +540,16 @@ _link (void)
 }
 
 int
-_unlink (void)
+_unlink (const char *path)
 {
+#ifdef ARM_RDI_MONITOR
+  int block[2];
+  block[0] = path;
+  block[1] = strlen(path);
+  return wrap (do_AngelSWI (AngelSWI_Reason_Remove, block)) ? -1 : 0;
+#else  
   return -1;
+#endif
 }
 
 void
@@ -546,9 +559,9 @@ _raise (void)
 }
 
 int
-_gettimeofday (struct timeval * tp, struct timezone * tzp)
+_gettimeofday (struct timeval * tp, void * tzvp)
 {
-
+  struct timezone *tzp = tzvp;
   if (tp)
     {
     /* Ask the host for the seconds since the Unix epoch.  */
@@ -599,24 +612,62 @@ _times (struct tms * tp)
 
 
 int
-isatty (int fd)
+_isatty (int fd)
 {
-  return 1;
-  fd = fd;
+#ifdef ARM_RDI_MONITOR
+  int fh = remap_handle (fd);
+  return wrap (do_AngelSWI (AngelSWI_Reason_IsTTY, &fh));
+#else
+  return (fd <= 2) ? 1 : 0;  /* one of stdin, stdout, stderr */
+#endif
 }
 
 int
 _system (const char *s)
 {
+#ifdef ARM_RDI_MONITOR
+  int block[2];
+  int e;
+
+  /* Hmmm.  The ARM debug interface specification doesn't say whether
+     SYS_SYSTEM does the right thing with a null argument, or assign any
+     meaning to its return value.  Try to do something reasonable....  */
+  if (!s)
+    return 1;  /* maybe there is a shell available? we can hope. :-P */
+  block[0] = s;
+  block[1] = strlen (s);
+  e = wrap (do_AngelSWI (AngelSWI_Reason_System, block));
+  if ((e >= 0) && (e < 256))
+    {
+      /* We have to convert e, an exit status to the encoded status of
+         the command.  To avoid hard coding the exit status, we simply
+	 loop until we find the right position.  */
+      int exit_code;
+
+      for (exit_code = e; e && WEXITSTATUS (e) != exit_code; e <<= 1)
+	continue;
+    }
+  return e;
+#else
   if (s == NULL)
     return 0;
   errno = ENOSYS;
   return -1;
+#endif
 }
 
 int
 _rename (const char * oldpath, const char * newpath)
 {
+#ifdef ARM_RDI_MONITOR
+  int block[4];
+  block[0] = oldpath;
+  block[1] = strlen(oldpath);
+  block[2] = newpath;
+  block[3] = strlen(newpath);
+  return wrap (do_AngelSWI (AngelSWI_Reason_Rename, block)) ? -1 : 0;
+#else  
   errno = ENOSYS;
   return -1;
+#endif
 }
