@@ -7,8 +7,9 @@
 #define _SYS_DIRENT_H
 
 #include <sys/types.h>
-#include <linux/dirent.h>
+#include <bits/dirent.h>
 #define _LIBC 1
+#define  NOT_IN_libc 1
 #include <sys/lock.h>
 #undef _LIBC
 
@@ -40,6 +41,7 @@ int closedir(DIR *);
 
 /* internal prototype */
 void _seekdir(DIR *dir,off_t offset);
+DIR *_opendir(const char *);
 
 #ifndef _POSIX_SOURCE
 long telldir (DIR *);
